/**
 * The MySensors Arduino library handles the wireless radio link and protocol
 * between your home built sensors/actuators and HA controller of choice.
 * The sensors forms a self healing radio network with optional repeaters. Each
 * repeater and gateway builds a routing tables in EEPROM which keeps track of the
 * network topology allowing messages to be routed to nodes.
 *
 * Created by Henrik Ekblad <henrik.ekblad@mysensors.org>
 * Copyright (C) 2013-2015 Sensnology AB
 * Full contributor list: https://github.com/mysensors/Arduino/graphs/contributors
 *
 * Documentation: http://www.mysensors.org
 * Support Forum: http://forum.mysensors.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 *******************************
 *
 * DESCRIPTION
 *
 * Example sketch showing how to send in DS1820B OneWire temperature readings back to the controller
 * http://www.mysensors.org/build/temp
 */

#include <MySensor.h>  
#include <SPI.h>
#include <DallasTemperature.h>
#include <OneWire.h>
#include <EEPROM.h>
#include <Wire.h>
#include <PinChangeInt.h>
//#include <PinChangeIntConfig.h>
#include "TimerOne.h"

#define DEBUG 1

#define COMPARE_TEMP 0 // Send temperature only if changed? 1 = Yes 0 = No
#define LOST_ACK_INT_MODE 5 //lost ack before switching to internal mode and the use of knob temperature
#define KNOB_RANGE_MAX 30
#define KNOB_RANGE_MIN 5
#define UPDATE_INTERVAL_MIN 1 //minimum update interval
#define UPDATE_INTERVAL_MAX 3600 //maximum update intervall
#define HYSTERESIS_RANGE_MIN 0.1
#define HYSTERESIS_RANGE_MAX 5
#define SETPOINT_RANGE_MIN -85
#define SETPOINT_RANGE_MAX 85

//External Pin configuration
#define ONE_WIRE_BUS 5 // Pin where dallas sensor is connected 
#define ZERO_CROSS 3  // Pin for zero cross detection
#define VALVE 6  // Pin for controling valve
#define GREEN_LED 4   // Pin for LED control
#define KNOB_TEMP 0   //(A0)Pin for analog input from knob potentiometer
#define BUTTON 17//Button pin for EEPROM erase

//External pin states
#define LED_ON 0
#define LED_OFF 1
#define VALVE_ON 1
#define VALVE_OFF 0
#define BUTTON_PRESSED 0

//EEPROM ADRESS definition
//Check the start adress EEPROM_LOCAL_CONFIG_ADDRESS before compilation, should be valid, but can eventually change with updates in mysensor libraries
#define DEFAULT_PARAM_EEPROM_WRITE_ADDR EEPROM_LOCAL_CONFIG_ADDRESS   //Address from Mysensors.h
#define SETPOINT_TEMP_ADDR (DEFAULT_PARAM_EEPROM_WRITE_ADDR+1)
#define HYST_LOW_ADDR (SETPOINT_TEMP_ADDR+4)
#define HYST_HIGH_ADDR (HYST_LOW_ADDR + 4)
#define UPDATE_INT_SEC_ADDR (HYST_HIGH_ADDR + 4)

//unsigned long SLEEP_TIME = 10000; // Sleep time between reads (in milliseconds)
OneWire oneWire(ONE_WIRE_BUS); // Setup a oneWire instance to communicate with any OneWire devices (not just Maxim/Dallas temperature ICs)
DallasTemperature sensors(&oneWire); // Pass the oneWire reference to Dallas Temperature. 

//MySensor specific
MySensor gw;

boolean receivedConfig = false;
boolean metric = true; 
// Initialize temperature message

MyMessage msgTemp(0,V_TEMP);
MyMessage msgFlowState(0,V_HVAC_FLOW_STATE);
MyMessage msgFlowMode(0,V_HVAC_FLOW_MODE);
MyMessage msgSetpointHeat(0,V_HVAC_SETPOINT_HEAT);
MyMessage msgknobTemperature(0,V_VAR1);
MyMessage msgHystLow(0,V_VAR3);
MyMessage msgHystHigh(0,V_VAR4);
MyMessage msgUpdateTime(0,V_VAR5);

//default values used at reset or at first start
float setpointTempDefault = 20;  //requested temperature sent to thermostat
float hysteresisHighDefault = 0.5;
float hysteresisLowDefault = 0.5;
unsigned long updateIntervalSecDefault = 10;


//Remote changeable values
float setpointTemp = 20;  //requested temperature sent to thermostat
float setpointTempIn = 0;
float setTemp = 27.3;              //temperature used in thermostat for regulation loop
float hysteresisHigh = 0.5;
float hysteresisHighIn = 0;
float hysteresisLow = 0.5;
float hysteresisLowIn = 0;
unsigned long updateIntervalSec = 10;         //requested sleepTime in seconds
unsigned long updateIntervalSecIn = 0;

float temperature = 0;
bool internalMode = true;

static volatile int countLostAck = 0;
static volatile bool receivedParameterUpdate = false;


static volatile int countButtonTimer = 0; //time depends on intitalized value of Timer1 interrupt
static volatile int countUpdateIntervalTimer = 0; //time depends on intitalized value of Timer1 interrupt



//Manaually direct changeable values (knob at thermostat)
float voMeasured = 0;
float knobTemperature = 10;    //requested temp set manually on thermostat
float knobTemperatureOld = 10;

void setup()  
{ 
  // Startup up the OneWire library
  sensors.begin();
  // requestTemperatures() will not block current thread
  sensors.setWaitForConversion(false);

  // Startup and initialize MySensors library. Set callback for incoming messages. 
  //gw.begin();
  gw.begin(incomingMessage, AUTO, true);

  digitalWrite(ZERO_CROSS, HIGH);
  pinMode(ZERO_CROSS, INPUT);

  digitalWrite(VALVE, VALVE_OFF);
  pinMode(VALVE, OUTPUT);
  digitalWrite(GREEN_LED, LED_OFF);
  pinMode(GREEN_LED, OUTPUT);

  //define the button pin
  pinMode(BUTTON, INPUT);
  digitalWrite(BUTTON,HIGH);

  //Initialize interrupt on button pin
  PCintPort::attachInterrupt(BUTTON, buttonInterrupt,CHANGE); // attach a PinChange Interrupt t

  // Send the sketch version information to the gateway and Controller
  gw.sendSketchInfo("Thermostat", "1.0");

  // Present sensors to controller 
     gw.present(0, S_HVAC);


  //writes default parameters to eeprom and reads user set parameters from EEPROM
  readWriteDefaultParametersEEPROM();
  //Initialize variables depending on default/user set parameters
  intializeVariables();
  //Set Timer to Value for to execute update on start up
  countUpdateIntervalTimer = updateIntervalSec;

    //This timer affect second count for button and also update interval
    Timer1.initialize(1000000);    //1 seconds
    Timer1.attachInterrupt(timer1Callback);
}

void loop()     
{     
  // Process incoming messages (like config from server)
  gw.process(); 

  //Checking knob temerature and if changed more than 0.3 degrees then send updated value to controller
  knobTemperature = readknobTemperature();
  if (knobTemperature<= knobTemperatureOld-0.3 | knobTemperature >= knobTemperatureOld +0.3){
     sendKnobTemperature(knobTemperature);
     knobTemperatureOld = knobTemperature;
     Serial.println("KNOB TEMPERATURE CHANGED");
     //Check if thermostat is intrernal mode if so the thermostat shall update
     if (internalMode == true){
      Serial.println("Internalmode = true");
      receivedParameterUpdate = true;
     }
  }

  //Check if update interval is reached or if a new parameter is received
  if(countUpdateIntervalTimer >= updateIntervalSec-1 | receivedParameterUpdate == true){
    //Resetting interval counter
    countUpdateIntervalTimer =0;
    //Setting received parameter update to cleared
    receivedParameterUpdate = false;
  
  
    if(DEBUG) Serial.println("--------------------------------");
    if(DEBUG) Serial.println("Starting thermostat update cycle");
    
    //read temperature value and return it
    temperature = readTemperature();
    //send temp value if value is bananas this indicates brooken temp sensor
    sendTemperature(temperature);

    //read and convert knob value and return corresponding temperature 
    knobTemperature = readknobTemperature(); 
    //Sending knob temperature to system
    sendKnobTemperature(knobTemperature);

    //Set Internal/External mode
    //assign value to set point either knobTemp as set point or received setPoint from controller
    if (countLostAck >= LOST_ACK_INT_MODE){
      //make sure that counter variable dont overflows and restarts at zero => External mode without ack received
      countLostAck = LOST_ACK_INT_MODE; 

      setTemp = knobTemperature;
      gw.send(msgFlowMode.set("Internal"));
      internalMode = true;
    } else{
      setTemp=setpointTemp;
      gw.send(msgFlowMode.set("External"));
      internalMode = false;
    }

  
 
    //Check that the sensor is within temperature range
    if (temperature != -127.00 && temperature != 85.00) {  
      //Check temperature and open/close valve
      if (temperature <= (setTemp-hysteresisLow)){
        //open valve
        digitalWrite(VALVE, VALVE_ON);
        digitalWrite(GREEN_LED, LED_ON);
        gw.send(msgFlowState.set("HeatOn"));
      }
      if (temperature >= (setTemp+hysteresisHigh)){
        //close valve
        digitalWrite(VALVE, VALVE_OFF);
        digitalWrite(GREEN_LED, LED_OFF);
        gw.send(msgFlowState.set("Off"));
      }   
    }  


    //Send setpoint temperature used for controling the thermostat
    sendSetPointHeat(setpointTemp); 
    sendHystLow(hysteresisLow);
    sendHystHigh(hysteresisHigh);
    sendUpdateInterval(updateIntervalSec);

    if(DEBUG) Serial.println("Cycle complete");
    if(DEBUG) Serial.println("--------------------------------");
  }

}

void readWriteDefaultParametersEEPROM(){
  //Writes default params to EEPROM at first startUp/reset and there after reads the values from EEPROM
  if(EEPROM.read(DEFAULT_PARAM_EEPROM_WRITE_ADDR) != 1){
    Serial.println("Nothing stored in EEPROM, writing default values");
    EEPROM.put(SETPOINT_TEMP_ADDR, setpointTempDefault);
    EEPROM.put(HYST_LOW_ADDR, hysteresisLowDefault);
    EEPROM.put(HYST_HIGH_ADDR, hysteresisHighDefault);
    EEPROM.put(UPDATE_INT_SEC_ADDR, updateIntervalSecDefault);
    EEPROM.put(DEFAULT_PARAM_EEPROM_WRITE_ADDR, 1);   //Marking in EEPROM that default params are written 
    Serial.println("EEPROM default values written");
  }else{
    Serial.println("Reading default values from EEPROM");
    EEPROM.get(SETPOINT_TEMP_ADDR, setpointTemp);
    EEPROM.get(HYST_LOW_ADDR, hysteresisLow);
    EEPROM.get(HYST_HIGH_ADDR, hysteresisHigh);
    EEPROM.get(UPDATE_INT_SEC_ADDR, updateIntervalSec);
    Serial.println("Done reading default values from EEPROM");
    
  }
}

//This function initialize varaibles dependant on default parameters
void intializeVariables(){
  //Initialize update interval
}

void buttonInterrupt(){
  //Serial.println("Button interrupt");
  if (digitalRead(BUTTON)==BUTTON_PRESSED){
    if (DEBUG) Serial.println("Button pressed");
    countButtonTimer = 0;
  }
  else{
    if (DEBUG) Serial.println("Button released");
    if (DEBUG) Serial.println(countButtonTimer);
    
    if (countButtonTimer >= 5){ // 5 seconds elapsed
      if (DEBUG) Serial.println("reset initiated");
      clearEeprom(); //clear EEPROM
      if (DEBUG) Serial.println("reset performed");
      readWriteDefaultParametersEEPROM(); //writes back default parameters to EEPROM
      intializeVariables(); //Initialize variables that needs after reset or in virgin mode
    }
  }
}

void timer1Callback(){
  countButtonTimer++;
  countUpdateIntervalTimer++;
  //Serial.print("countupdate interval timer: ");
  //Serial.println(countUpdateIntervalTimer);
}


/*send messages*/
/*-----------------------------------------------------------------------*/
void sendTemperature(float sendTemp){
  gw.send(msgTemp.setSensor(0).set(temperature,1), true); //message including request for acknowledge
  countLostAck++; //count up lost acknowledge. Reset is in incoming isAck check
}

void sendSetPointHeat(float setPointTmp)
{
   gw.send(msgSetpointHeat.set(setPointTmp,1));
}

void sendKnobTemperature(float knobTempTmp){
  gw.send(msgknobTemperature.set(knobTempTmp,1));
}

void sendFlowMode(String flowModeMsg){
  //gw.send(msgFlowMode.set(flowModeMsg));
  Serial.println("Sending flow mode");
  Serial.println(flowModeMsg);
}

void sendHystLow(float hysteresisLowTmp){
  gw.send(msgHystLow.set(hysteresisLowTmp,1));
}

void sendHystHigh(float hysteresisHighTmp){
  gw.send(msgHystHigh.set(hysteresisHighTmp,1));
}

void sendUpdateInterval(unsigned long updateIntervalTmp){
  gw.send(msgUpdateTime.set(updateIntervalTmp));
}
/*-----------------------------------------------------------------------*/

//read and convert knob value and return corresponding temperature
float readknobTemperature(){
   uint16_t voMeasured = analogRead(KNOB_TEMP);// Get Knob temperature value
   float knobTemp = 0;
   int maxAdcValue = 1023; 
   
   knobTemp = voMeasured*(KNOB_RANGE_MAX-KNOB_RANGE_MIN)/maxAdcValue + KNOB_RANGE_MIN;

   return knobTemp;
}


float readTemperature(){
  
  // Fetch temperatures from Dallas sensors
  sensors.requestTemperatures();

  // query conversion time and sleep until conversion completed
  int16_t conversionTime = sensors.millisToWaitForConversion(sensors.getResolution());
  // sleep() call can be replaced by wait() call if node need to process incoming messages (or if node is repeater)
  gw.wait(conversionTime);

  // Read temperatures and send them to controller  
  // Fetch and round temperature to one decimal
  float readTemp = static_cast<float>(static_cast<int>((gw.getConfig().isMetric?sensors.getTempCByIndex(0):sensors.getTempFByIndex(0)) * 10.)) / 10.;
  return readTemp;
}


void clearEeprom(){
  /***
    Iterate through each byte of the EEPROM storage.

    Larger AVR processors have larger EEPROM sizes, E.g:
    - Arduno Duemilanove: 512b EEPROM storage.
    - Arduino Uno:        1kb EEPROM storage.
    - Arduino Mega:       4kb EEPROM storage.

    Rather than hard-coding the length, you should use the pre-provided length function.
    This will make your code portable to all AVR processors.
  ***/

  for (int i = 0 ; i < EEPROM.length() ; i++) {
    EEPROM.write(i, 255);
  }
  
}

void incomingMessage(const MyMessage &message) {    
    Serial.println("Receiving message");
    String messageTmp = message.getString();

    //Check if message is Ack and then reset lost ack Counter
    if(message.isAck()){
      if (DEBUG){
        Serial.println("Ack received");
        Serial.print("lost acknowledge: ");
        Serial.println(countLostAck-1);
      }
      countLostAck = 0;
      
    }else if (messageTmp == ""){
      if (DEBUG) Serial.println("Message empty");
    }else{
      if (DEBUG) Serial.println(messageTmp);
    }
    
    // We only expect one type of message from controller. But we better check anyway.
    if (message.type==V_HVAC_SETPOINT_HEAT) {
      Serial.println("Setpoint heat reaceived");
      //setpointTempIn = (float) atof(message.getString());
      setpointTempIn = message.getFloat();
      Serial.println(setpointTempIn);
      if (setpointTempIn >= SETPOINT_RANGE_MIN && setpointTempIn <= SETPOINT_RANGE_MAX){
        setpointTemp = setpointTempIn;
        receivedParameterUpdate = true;  //tell main loop that new parameter is received
        //Updating dafault parameters
        EEPROM.put(SETPOINT_TEMP_ADDR, setpointTemp);
      }
      //return message to controller
      sendSetPointHeat(setpointTemp);
    }

    if (message.type==V_VAR3) {
       //hysteresisLowIn = (float) atof(message.getString());
       hysteresisLowIn = message.getFloat();
       if (hysteresisLowIn >= HYSTERESIS_RANGE_MIN && hysteresisLowIn <= HYSTERESIS_RANGE_MAX){
        hysteresisLow = hysteresisLowIn;
        receivedParameterUpdate = true;  //tell main loop that new parameter is received
        //Updating dafault parameters
        EEPROM.put(HYST_LOW_ADDR, hysteresisLow);
       }
       //Send value that is in thermostat even if incoming message was out of range 
       sendHystLow(hysteresisLow);
    }
    
    if (message.type==V_VAR4) {
       //hysteresisHighIn = (float) atof(message.getString());
       hysteresisHighIn = message.getFloat();
       if (hysteresisHighIn >= HYSTERESIS_RANGE_MIN && hysteresisHighIn <= HYSTERESIS_RANGE_MAX){
        hysteresisHigh = hysteresisHighIn;
        receivedParameterUpdate = true;  //tell main loop that new parameter is received
        //Updating dafault parameters
        EEPROM.put(HYST_HIGH_ADDR, hysteresisHigh);
       }
       //Send value that is in thermostat even if incoming message was out of range 
       sendHystHigh(hysteresisHigh);
    }
    
    if (message.type==V_VAR5) {
      //updateIntervalSec = (unsigned long) atol(message.getString());
      updateIntervalSecIn = message.getLong();
      //Check interval should be 1 (1s) < interval < 3600 (1h)
      if (updateIntervalSecIn >= UPDATE_INTERVAL_MIN && updateIntervalSecIn <=UPDATE_INTERVAL_MAX){
        updateIntervalSec = updateIntervalSecIn;
        receivedParameterUpdate = true;  //tell main loop that new parameter is received
        //Updating dafault parameters
        EEPROM.put(UPDATE_INT_SEC_ADDR, updateIntervalSecIn);
      }
      sendUpdateInterval(updateIntervalSec);
    }
}











